/**
  *  \file ptrmultilist.cc
  *  \brief Multiple-Iterator Pointer List
  *
  *  This class is implemented using almost doubly-linked lists:
  *  - a list of elements
  *  - a list of live iterators
  *
  *  The almost doubly-linked lists stores a pointer to the next element (next_iterator),
  *  as well as a pointer to the pointer to this element (pthis_iterator). This allows
  *  easy prepend, append, and removal of elements, but no backward traversal.
  *
  *  Each iterator contains a pointer to the current element as well as one to the next one.
  *  When an element is deleted, we can thus simply keep a reference into the list.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/ptrmultilist.h"
#include "cpluslib/assert.h"

/** Assign iterator. */
ptr_multi_list_base::iterator&
ptr_multi_list_base::iterator::operator=(const ptr_multi_list_base::iterator& other)
{
    if (this != &other) {
        unlink();
        item = other.item;
        next_item = other.next_item;
        link(other);
    }
    return *this;
}

/** Internal: unlink this iterator. It will no longer be associated with a list. */
void
ptr_multi_list_base::iterator::unlink()
{
    if (pthis_iterator != 0) {
        // Unlink
        ASSERT(*pthis_iterator == this);
        *pthis_iterator = next_iterator;
        if (next_iterator != 0) {
            next_iterator->pthis_iterator = pthis_iterator;
        }

        // Make operation idempotent
        pthis_iterator = 0;
        next_iterator = 0;
    }
}

/** Internal: link this iterator. Associates it with the same list as other. */
void
ptr_multi_list_base::iterator::link(const iterator& other)
{
    if (other.pthis_iterator != 0) {
        // The other one is a live iterator
        next_iterator = const_cast<iterator*>(&other);
        pthis_iterator = other.pthis_iterator;
        *pthis_iterator = this;
        other.pthis_iterator = &next_iterator;
    } else {
        // The other one is not a live iterator
        next_iterator = 0;
        pthis_iterator = 0;
    }
}

/** Clear list. Deletes all elements. */
void
ptr_multi_list_base::clear()
{
    // First invalidate iterators, to save work later on erase()
    invalidate_all_iterators();

    // Clear the container. This could actually be implemented even simpler.
    while (!empty()) {
        deleter(extract_front());
    }
}

/** Prepend an element.
    If appending fails, deletes the element before rethrowing the exception. */
void
ptr_multi_list_base::push_front(void* data)
{
    try {
        node* n = new node;
        n->data = data;
        n->next_node = first_node;
        n->pthis_node = &first_node;
        if (first_node == 0) {
            plast_node = &n->next_node;
        }
        first_node = n;
    }
    catch (...) {
        deleter(data);
        throw;
    }
}

/** Erase element. Deletes it in the process. */
void
ptr_multi_list_base::erase(const iterator& iter)
{
    deleter(extract(iter));
}

/** Insert element before. */
ptr_multi_list_base::iterator
ptr_multi_list_base::insert(const iterator& iter, void* data)
{
    node* result;
    if (iter.item == 0) {
        if (iter.next_item == 0) {
            // Case 1: iterator points at end, so this is push_back.
            result = insert_end(data);
        } else {
            // Case 2: iterator points at a deleted element.
            // Insert before next_item.
            result = insert_before(iter.next_item, data);
        }
    } else {
        // Case 3: iterator is valid. Insert before item.
        result = insert_before(iter.item, data);
    }
    return iterator(result, &first_iterator);
}

/** Extract element. Removes an element from the sequence and returns it. */
void*
ptr_multi_list_base::extract(const iterator& iter)
{
    void* result;
    node*const item = iter.item;
    if (item != 0) {
        // Erase iterators
        for (iterator* i = first_iterator; i != 0; i = i->next_iterator) {
            if (item == i->item) {
                i->item = 0;
            } else if (item == i->next_item) {
                i->next_item = i->next_item->next_node;
            }
        }

        // Delete item
        *item->pthis_node = item->next_node;
        if (item->next_node) {
            item->next_node->pthis_node = item->pthis_node;
        } else {
            plast_node = item->pthis_node;
        }
        result = item->data;
        delete item;
    } else {
        result = 0;
    }
    return result;
}

/** Invalidate all iterators. Turns them all into end() iterators. */
void
ptr_multi_list_base::invalidate_all_iterators()
{
    while (first_iterator != 0) {
        first_iterator->invalidate();
    }
}

/** Insert a new element at the end. */
ptr_multi_list_base::node*
ptr_multi_list_base::insert_end(void* data)
{
    try {
        node* n = new node;
        n->data = data;
        n->next_node = 0;
        n->pthis_node = plast_node;
        *plast_node = n;
        plast_node = &n->next_node;
        return n;
    }
    catch (...) {
        deleter(data);
        throw;
    }
}

/** Insert a new element before an item. */
ptr_multi_list_base::node*
ptr_multi_list_base::insert_before(node* item, void* data)
{
    ASSERT(item != 0);
    try {
        // Insert new node
        node* n = new node;
        n->data = data;
        n->next_node = item;
        n->pthis_node = item->pthis_node;
        *n->pthis_node = n;
        item->pthis_node = &n->next_node;

        // Rewrite iterators
        // Do not rewrite iterators pointing at a deleted item.
        // In particular, this could be the iterator used to insert the element.
        for (iterator* i = first_iterator; i != 0; i = i->next_iterator) {
            if (i->item != 0 && i->next_item == item) {
                i->next_item = n;
            }
        }
        return n;
    }
    catch (...) {
        deleter(data);
        throw;
    }
}

/** Get size of list. */
std::size_t
ptr_multi_list_base::size() const
{
    std::size_t result = 0;
    for (node* n = first_node; n != 0; n = n->next_node) {
        ++result;
    }
    return result;
}
